#!/usr/bin/env python

#
#Copyrigt (c) 2008
#		Matt Smith <melkor@orangepalantir.org>
#
#This file is part of GraphCutie,This file is part of Foobar.
#
#GraphCutie is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#GraphCutie is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with GraphCutie, in the file COPYING.  
#If not, see <http://www.gnu.org/licenses/>.
#
#GraphCutie uses both python and PyQT, GPL version.  The copyright notices
#can be found in copyright_notice.txt

"""
Main calculator module, it contains the calculator class which can be used independantly of the Qt environment.

"""

import operator as o
from copy import deepcopy

from math import *
import time
VARIABLES = {	'pi':pi,
			'e':e}
FUNCTIONS = {	'sin':sin,
			'cos':cos,
			'tan':tan,
			'asin':asin,
			'acos':acos,
			'atan':atan,
			'exp':exp,
			'fabs':fabs,
			'ln':log,
			'log':log10,
			'sinh':sinh,
			'cosh':cosh,
			'tanh':tanh,
			'sqrt':sqrt
			}

OPERATORS = {'/':o.truediv,'*':o.mul,'+':o.add,'-':o.sub,'^':o.pow,'=':None}
INTS = ['1','2','3','4','5','6','7','8','9','0','.']
SPECIALS = INTS
SPECIALS.extend(OPERATORS.keys())
PM = {'+':o.pos,
	'-':o.neg}
from Parse import Parse
from Evaluate import Evaluate
	
class Calculator(object):
	"""This class is the simplest way to access the parse, and evaluate tools"""
	def __init__(self):
		self.precision = 6
		self.statement = "%s = %.*f"
		self.currentfunction = lambda x: 0
	def __call__(self,x):
		"""this is a convenience function that returns numbers.  You have to handle the exceptions yourself."""
		return self.currentfunction(x)
	def set(self,newfunction):
		"""This creates a new function and makes it the main function to be used with __call__ it also does not handle errors and will raise exceptions from invalid user data"""
		op_list = Parse(newfunction)
		check,fname = StatMakeFunction(op_list)
		if check==True:
			text = "User Defined Function: %s"%newfunction
			FUNCTIONS.get(op_list[0]).__doc__ = text
		self.currentfunction = FUNCTIONS.get(op_list[0])
		return check,op_list[0]
	def FromExpression(self,rawstring):
		"""take an expression and returns a tuple, the first element of the tuple denotes the success
		the second element of the tuple is a response string.  The variable 'ans' is assigned the value
		to a expression without an assignment."""
		try:
			op_list = Parse(rawstring)
		except SyntaxError, e:
			return False,str(e)
		try:
			if '=' in op_list and op_list[1]=='=':
				if op_list[0][0] in SPECIALS:
					return False,"Illegal assignment, %s cannot be used as a variable"%op_list[0]
				value = Evaluate(op_list[2:])
				VARIABLES[op_list[0]] = value
				return True,self.statement%(op_list[0],self.precision,value)
			elif '=' in op_list and op_list[2]=='=':
				if op_list[0][0] in SPECIALS:
					return False,"Illegal assignment, %s cannot be used as a function name"%op_list[0]
				check,response = StatMakeFunction(op_list)
				if check==True:
					text = "User Defined Function: %s"%rawstring
					FUNCTIONS.get(op_list[0]).__doc__ = text
				return True,rawstring
			else:
				value = Evaluate(op_list)
				VARIABLES['ans'] = value
				return True,self.statement%('ans',self.precision,value)
		except ValueError, e:
			return False,"%s is not defined."%str(e)
		except IndexError, e:
			return False,"Trailing operator %s"%str(e)
		except ZeroDivisionError, e:
			return False, "Zero Division Error %s"%str(e)
	def SetPrecision(self,nv):
		"""Sets the number of decimal places retained in a calculation"""
		self.precision = nv
	def SetDType(self,dt):
		"""Changes the output from decimal to scientific notation."""
		if dt==0:
			self.statement = "%s = %.*f"
		else:
			self.statement = "%s = %.*e"
	
def StatMakeFunction(op_list):
	"""This will create a function from the expression list then it will put the value into the FUNCTIONS dictionary in which case it can be executed"""
	try:
		fname,varname,expression = op_list[0],op_list[1],op_list[3:]
	except:
		return False,"No Function"
	tvn = "xxTEMP%i"%int(time.time())
	DeepSwap(expression,varname[0],tvn)
	def newfunction(x,arg=tvn):
		"""User defined function"""
		VARIABLES[arg]= x
		mycopy = deepcopy(expression)
		return Evaluate(mycopy)
	FUNCTIONS[fname]=newfunction
	return True,"%s has been defined"%fname

def DeepSwap(el,v,tvn):
	"""This is to swap out all of the arguments that are the arguments of a function."""
	for i in xrange(len(el)):
		if type(el[i]) == list:
			DeepSwap(el[i],v,tvn)
		elif el[i]==v:
			el[i]=tvn